const coral = @import("coral");

const kym = @import("../kym.zig");

associative: AssociativeTable,
contiguous: ContiguousList,

// TODO: Modify hash traits to be support fat contexts rather than passing Envs into the table.
const AssociativeTable = coral.map.Table(Field, *kym.RuntimeRef, .{
	.hash = Field.hash,
	.match = Field.match,
});

const ContiguousList = coral.list.Stack(?*kym.RuntimeRef);

const Field = struct {
	env: *kym.RuntimeEnv,
	ref: *kym.RuntimeRef,

	fn hash(field: Field) usize {
		return switch (field.env.unbox(field.ref)) {

		};
	}

	fn match(self: Field, other: Field) bool {
		return kym.test_equality(self.env, self.ref, other.ref);
	}
};

const Self = @This();

pub fn new(env: *kym.RuntimeEnv) kym.RuntimeError!*kym.RuntimeRef {
	var self = Self{
		.associative = AssociativeTable.make(env.allocator),
		.contiguous = ContiguousList.make(env.allocator),
	};

	errdefer {
		self.associative.free();
		self.contiguous.free();
	}

	return try env.new_dynamic(coral.io.bytes_of(&self), &typeinfo);
}

const typeinfo = kym.Typeinfo{
	.name = "table",
	.clean = typeinfo_clean,
	.get = typeinfo_get,
	.set = typeinfo_set,
};

fn typeinfo_clean(context: kym.Context) void {
	const table = @as(*Self, @ptrCast(@alignCast(context.userdata.ptr)));

	{
		var field_iterable = table.associative.as_iterable();

		while (field_iterable.next()) |entry| {
			context.env.discard(entry.key.ref);
			context.env.discard(entry.value);
		}
	}

	table.associative.free();

	while (table.contiguous.pop()) |ref| {
		context.env.discard(ref);
	}

	table.contiguous.free();
}

fn typeinfo_get(context: kym.Context, index_ref: *const kym.RuntimeRef) kym.RuntimeError!?*kym.RuntimeRef {
	const table = @as(*Self, @ptrCast(@alignCast(context.userdata.ptr)));

	switch (context.env.unbox(index_ref)) {
		.symbol, .string => {
			return context.env.acquire(table.associative.lookup(index_ref));
		},

		.number => |number| {
			// TODO: Implement dedicated integer type within VM internals for this sort of indexing.
			if (@trunc(number) == number) {
				const index = @as(usize, @intFromFloat(number));

				if (index < table.contiguous.values.len) {
					return context.env.acquire(table.contiguous.values[index] orelse return null);
				}
			}

			return context.env.acquire(table.associative.lookup(.{
				.env = context.env,
				.ref = index_ref,
			}));
		},

		else => return context.env.raise(error.TypeMismatch, "expected symbol, string, or number index"),
	}
}

fn typeinfo_set(context: kym.Context, index_ref: *const kym.RuntimeRef, value_ref: ?*const kym.RuntimeRef) kym.RuntimeError!void {
	const table = @as(*Self, @ptrCast(@alignCast(context.userdata.ptr)));

	switch (context.env.unbox(context.index_ref)) {
		.symbol, .string => {
			const acquired_index_ref = context.env.acquire(index_ref);

			errdefer context.env.discard(acquired_index_ref);

			const acquired_value_ref = context.env.acquire(value_ref);

			errdefer context.env.discard(acquired_value_ref);

			if (try table.associative.replace(acquired_index_ref, acquired_value_ref)) |replaced| {
				context.env.discard(replaced.key);
				context.env.discard(replaced.value);
			}
		},

		.number => |number| {
			const index = coral.math.clamped_cast(@typeInfo(usize), number);

			if (index == number) {
				if (index < table.contiguous.values.len) {
					table.contiguous.values[index] = context.env.acquire(value_ref);

					return;
				}

				if (index == table.contiguous.values.len) {
					const acquired_index_ref = context.env.acquire(value_ref);

					errdefer context.env.discard(acquired_index_ref);

					try table.contiguous.push_one(acquired_index_ref);

					return;
				}
			}

			const acquired_index_ref = context.env.acquire(index_ref);

			errdefer context.env.discard(index_ref);

			const acquired_value_ref = context.env.acquire(value_ref);

			errdefer context.env.discard(value_ref);

			if (try table.associative.replace(acquired_index_ref, acquired_value_ref)) |replaced| {
				context.env.discard(replaced.key);
				context.env.discard(replaced.value);
			}
		},

		else => return context.env.raise(error.TypeMismatch, "expected symbol, string, or number index"),
	}
}
